@@ -21,6 +21,7 @@ from pprint import pformat as pfmt
 from typing import Callable, List, Optional, Tuple
 
 from lightning.fabric.utilities import rank_zero_info
+from lightning.fabric.utilities.imports import _TORCH_GREATER_EQUAL_1_13
 from lightning.fabric.utilities.types import ReduceLROnPlateau
 from lightning.pytorch import LightningModule, Trainer
 from lightning.pytorch.callbacks import Callback
@@ -88,6 +89,15 @@ class StrategyAdapter:
         assert self.pl_module._trainer is not None
         return self.pl_module._trainer.strategy
 
+    @property
+    def using_sharded_optimizer(self) -> bool:
+        """Whether the currently used optimizer is a supported sharded optimizer.
+
+        Returns:
+            bool: Returns ``True`` if the current optimizer is a supported sharded optimizer.
+        """
+        return hasattr(self.pl_module.trainer.optimizers[0], "consolidate_state_dict")
+
     def on_before_init_fts(self) -> None:
         """Hook executed in :class:`~finetuning_scheduler.fts.FinetuningScheduler` setup immediately before
         :meth:`~finetuning_scheduler.fts_supporters.ScheduleImplMixin.init_fts`
@@ -115,9 +125,8 @@ class StrategyAdapter:
     def fts_optim_transform(self, orig_pl: List, inspect_only: bool = False) -> List:
         """A method that can be overridden by a :class:`~finetuning_scheduler.strategy_adapters.StrategyAdapter` if
         a :external+pl:class:`~lightning.pytorch.strategies.Strategy` performs parameter transformations that cause
-        the current :external+torch:class:`~torch.optim.Optimizer`'s view of parameter names to diverge from the
-        original parameter names. By default, no transformation of schedule parameter names is required for
-        optimizer operations.
+        the current optimizer's view of parameter names to diverge from the original parameter names. By default,
+        no transformation of schedule parameter names is required for optimizer operations.
 
         Args:
             orig_pl (List): The original parameter name list before a given
@@ -127,8 +136,7 @@ class StrategyAdapter:
                 inspection and validation contexts.
 
         Returns:
-            List: A transformed parameter name list that matches the current
-                :external+torch:class:`~torch.optim.Optimizer`'s view of them after a given
+            List: A transformed parameter name list that matches the current optimizer's view of them after a given
                 :external+pl:class:`~lightning.pytorch.strategies.Strategy`'s transformation of the original parameter
                 names.
         """
@@ -139,14 +147,13 @@ class StrategyAdapter:
         :meth:`~finetuning_scheduler.strategy_adapters.StrategyAdapter.fts_optim_transform`. Can be overridden by a
         :class:`~finetuning_scheduler.strategy_adapters.StrategyAdapter` if a
         :external+pl:class:`~lightning.pytorch.strategies.Strategy` performs parameter transformations that cause the
-        original user view of parameter names to diverge from the current
-        :external+torch:class:`~torch.optim.Optimizer`'s view. By default, no transformation of
-        :external+torch:class:`~torch.optim.Optimizer` parameter names is required.
+        original user view of parameter names to diverge from the current optimizer's view. By default, no
+        transformation of optimizer parameter names is required.
 
         Args:
-            param_names (List): A parameter name list from the current :external+torch:class:`~torch.optim.Optimizer`'s
-                view of them after a :external+pl:class:`~lightning.pytorch.strategies.Strategy`'s transformation of the
-                original parameter names.
+            param_names (List): A parameter name list from the current optimizer's view of them after a
+                :external+pl:class:`~lightning.pytorch.strategies.Strategy`'s transformation of the original parameter
+                names.
 
         Returns:
             List: The original parameter name list before a given
@@ -171,9 +178,51 @@ class StrategyAdapter:
             orig_num_pgs.append(len(optimizer.param_groups))
             optimizer.param_groups = []
         for lrs_cfg in trainer.lr_scheduler_configs:
-            lrs_cfg.scheduler.base_lrs = []
+            lrs_cfg.scheduler.last_epoch = -1  # type: ignore[union-attr]
+            if not isinstance(lrs_cfg.scheduler, ReduceLROnPlateau):
+                lrs_cfg.scheduler.base_lrs = []
         return orig_num_pgs
 
+    def _reconfigure_optimizer_for_phase0(self, trainer: Trainer) -> None:
+        """Reconfigure optimizer state to comport with the scheduled phase 0.
+
+        Args:
+            trainer (Trainer): The :external+pl:class:`~lightning.pytorch.trainer.trainer.Trainer` object.
+        """
+        if self.using_sharded_optimizer:
+            # update the parent (sharded) optimizer defaults with the wrapped optimizer defaults to ensure they are
+            # included in our newly configured param groups
+            trainer.optimizers[0].defaults.update(trainer.optimizers[0].optim.defaults)
+        # thaw only params scheduled in phase 0
+        self.fts_handle.step_pg(depth=self.fts_handle.curr_depth, optimizer=trainer.optimizers[0], depth_sync=False)
+        if self.using_sharded_optimizer:
+            # update internal sharded optimizer state with the new set of parameters
+            trainer.optimizers[0]._verify_and_init_params([p for p in self.pl_module.parameters() if p.requires_grad])
+            # repartition the sharded optimizer state
+            self.fts_handle._repartition_sharded_optim(trainer.optimizers[0])
+
+    def _reconfigure_lrs_for_phase0(self, trainer: Trainer, orig_num_pgs: List) -> None:
+        """Reconfigure lr scheduler state to comport with the scheduled phase 0.
+
+        Args:
+            trainer (Trainer): The :external+pl:class:`~lightning.pytorch.trainer.trainer.Trainer` object.
+            orig_num_pgs (List): A list of the number of parameter groups pruned for each optimizer (since only a single
+                optimizer is currently supported by FTS, this list will have only a single element in this verison.)
+        """
+        # since we may have added parameter groups (e.g. implementing ``no_decay`` for user), we need to reinitialize
+        # certain lr_scheduler variables (including type-dependent ones like ``min_lrs`` and ``lr_lambdas``)
+        if trainer.lr_scheduler_configs:
+            for lrs_cfg in trainer.lr_scheduler_configs:
+                if _TORCH_GREATER_EQUAL_1_13 and not isinstance(lrs_cfg.scheduler, ReduceLROnPlateau):
+                    lrs_cfg.scheduler._initial_step()
+                lrs_cfg.scheduler._last_lr = [  # type: ignore[union-attr]
+                    group["lr"] for group in lrs_cfg.scheduler.optimizer.param_groups
+                ]
+                if isinstance(lrs_cfg.scheduler, ReduceLROnPlateau):
+                    lrs_cfg.scheduler.min_lrs = lrs_cfg.scheduler.min_lrs[orig_num_pgs[0] :]
+                elif hasattr(lrs_cfg.scheduler, "lr_lambdas"):
+                    lrs_cfg.scheduler.lr_lambdas = lrs_cfg.scheduler.lr_lambdas[orig_num_pgs[0] :]
+
     def phase0_optimizer_override(self) -> None:
         """Reconfigure the user-configured optimizer (configured via `configure_optimizers`) to optimize the
         parameters (and only those parameters) scheduled to be optimized in phase 0 of the current fine-tuning
@@ -187,16 +236,8 @@ class StrategyAdapter:
         orig_num_pgs = StrategyAdapter._clean_optim_lr_pgs(trainer)
         # refreeze in case user has thawed parameters not present in phase 0
         self.fts_handle.freeze_before_training(self.pl_module)
-        # thaw only params scheduled in phase 0
-        self.fts_handle.step_pg(depth=self.fts_handle.curr_depth, optimizer=trainer.optimizers[0], depth_sync=False)
-        # since we may have added parameter groups (e.g. implementing ``no_decay`` for user), we need to reinitialize
-        # certain lr_scheduler variables (including type-dependent ones like ``min_lrs`` and ``lr_lambdas``)
-        for lrs_cfg in trainer.lr_scheduler_configs:
-            lrs_cfg.scheduler._last_lr = [group["lr"] for group in lrs_cfg.scheduler.optimizer.param_groups]
-            if isinstance(lrs_cfg.scheduler, ReduceLROnPlateau):
-                lrs_cfg.scheduler.min_lrs = lrs_cfg.scheduler.min_lrs[orig_num_pgs[0] :]
-            elif hasattr(lrs_cfg.scheduler, "lr_lambdas"):
-                lrs_cfg.scheduler.lr_lambdas = lrs_cfg.scheduler.lr_lambdas[orig_num_pgs[0] :]
+        self._reconfigure_optimizer_for_phase0(trainer)
+        self._reconfigure_lrs_for_phase0(trainer, orig_num_pgs)
         p0_override_msg = self.fts_handle.PHASE_0_DIVERGENCE_MSG + (
             "Since `enforce_phase0_params` is currently set to `True` (the default), FinetuningScheduler has"
             " reconfigured the optimizer to optimize the parameters (and only those parameters) scheduled to be"
