@@ -18,9 +18,13 @@ from unittest import mock
 
 import pytest
 import torch
-from lightning.fabric.utilities.imports import _TORCH_GREATER_EQUAL_1_13, _TORCH_GREATER_EQUAL_2_0
+from lightning.fabric.utilities.imports import (
+    _TORCH_GREATER_EQUAL_1_13,
+    _TORCH_GREATER_EQUAL_2_0,
+    _TORCH_GREATER_EQUAL_2_1,
+)
 from lightning.pytorch import seed_everything, Trainer
-from lightning.pytorch.plugins.precision.fsdp import FSDPMixedPrecisionPlugin
+from lightning.pytorch.plugins.precision.fsdp import FSDPPrecisionPlugin
 from lightning.pytorch.strategies import FSDPStrategy
 from lightning.pytorch.utilities.exceptions import MisconfigurationException
 from torch.utils.data import DataLoader
@@ -34,6 +38,7 @@ from tests.test_finetuning_scheduler_callback import (
     ExplicitLossFTSCheckpoint,
     FinetuningSchedulerBoringModel,
     get_fts,
+    nones,
     TestFinetuningScheduler,
 )
 
@@ -46,7 +51,7 @@ if _min_fsdp_available:
         checkpoint_wrapper,
         CheckpointImpl,
     )
-    from torch.distributed.fsdp.fully_sharded_data_parallel import CPUOffload, FullyShardedDataParallel
+    from torch.distributed.fsdp.fully_sharded_data_parallel import CPUOffload, FullyShardedDataParallel, MixedPrecision
     from torch.distributed.fsdp.wrap import size_based_auto_wrap_policy, wrap
 else:
     FullyShardedDataParallel = None  # type: ignore[misc,assignment]
@@ -56,27 +61,33 @@ else:
     size_based_auto_wrap_policy = object
     wrap = object
 
-if _TORCH_GREATER_EQUAL_2_0:
-    from torch.distributed.fsdp.wrap import _FSDPPolicy
-else:
+if _TORCH_GREATER_EQUAL_2_1:
+    from torch.distributed.fsdp.wrap import CustomPolicy
+
+    DISABLE_USE_ORIG = {"use_orig_params": False}
     _FSDPPolicy = object
+elif _TORCH_GREATER_EQUAL_2_0:
+    from torch.distributed.fsdp.wrap import _FSDPPolicy
+
+    DISABLE_USE_ORIG = {"use_orig_params": False}
+else:
+    CustomPolicy = object
+    _FSDPPolicy = object
+    DISABLE_USE_ORIG = {}
 
 
 additional_fsdp_warns = [
     "The number of training batches",  # minimizing cost of training for these tests
-    "The distutils package is deprecated",  # for tensorboard (but not tensorboardX) import as of PT 1.13.1
-    "`tensorboardX` has been removed as a depend",  # in case tensorboard/tensorboardX are not available
-    "is still running",  # subprocess is implicitly cleaned up
+    "does not support loading the optimizer",  # with PyTorch 1.x Lightning lacks OSD restoration support
     "Please use torch.distributed.all_gather_into_tensor",  # can be removed once PyTorch stops using internally,
     "Please use torch.distributed.reduce_scatter_tensor",  # can be removed once PyTorch stops using internally,
     "when logging on epoch level in distributed",  # validating FTS handling in this scenario
-    "Deallocating Tensor that still has live",  # TODO: investigate the occasional occurance of this warning
+    "Deallocating Tensor that still has live",  # still encounter this in some edge cases
 ]
 EXPECTED_WARNS.extend(additional_fsdp_warns)
 FSDP_BASE_WARNS = EXPECTED_WARNS
 FSDP_DYNAMO_EXPECTED_WARNS = [
     "Final phase max_transition_epoch",
-    "Your compiler for AOTAutograd is returning",  # out of initial scope
 ]
 
 ##########################
@@ -84,7 +95,7 @@ FSDP_DYNAMO_EXPECTED_WARNS = [
 ##########################
 
 
-@pytest.fixture(scope="function")
+@pytest.fixture(scope="module")
 def fsdp_ft_schedules(tmpdir_factory) -> Tuple[Path, Dict]:
     """Generates a default fine-tuning schedule for 'implicit' testing, a modified one for 'explicit' mode and an
     epoch-driven transitions only one for epoch_transitions_only testing."""
@@ -110,6 +121,46 @@ def fsdp_ft_schedules(tmpdir_factory) -> Tuple[Path, Dict]:
     fsdp_gen_sched_dict[0]["max_transition_epoch"] = 1
     fsdp_gen_sched_dict[1]["params"] = ["layer.[1-4].*"]
     fsdp_gen_sched_dict[1]["max_transition_epoch"] = 2
+    fsdp_reinit_optim_sched_dict = deepcopy(fsdp_gen_sched_dict)
+    fsdp_reinitlr_sched_dict = deepcopy(fsdp_gen_sched_dict)
+    fsdp_reinit_optim_sched_dict[1]["new_optimizer"] = {
+        "optimizer_init": {
+            "class_path": "torch.optim.Adam",
+            "init_args": {"lr": 2.1e-04},
+        },
+    }
+    fsdp_reinit_optim_sched_dict[2]["new_optimizer"] = {
+        "optimizer_init": {
+            "class_path": "torch.optim.SGD",
+            "init_args": {"lr": 2.0e-03, "momentum": 0.9, "weight_decay": 2.0e-06},
+        }
+    }
+    fsdp_reinitlr_sched_dict[1]["new_lr_scheduler"] = {
+        "lr_scheduler_init": {
+            "class_path": "torch.optim.lr_scheduler.StepLR",
+            "init_args": {"step_size": 1, "gamma": 0.7, "verbose": True},
+        },
+        "pl_lrs_cfg": {"interval": "epoch", "frequency": 1, "name": "Custom_Reinit_LR"},
+    }
+    fsdp_reinitlr_sched_dict[2]["lr"] = 3.0e-06
+    fsdp_reinitlr_sched_dict[2]["new_lr_scheduler"] = {
+        "lr_scheduler_init": {
+            "class_path": "torch.optim.lr_scheduler.CosineAnnealingWarmRestarts",
+            "init_args": {"T_0": 1, "T_mult": 2, "eta_min": 1.0e-07},
+        },
+        "pl_lrs_cfg": {"interval": "epoch", "frequency": 1, "name": "Custom_Reinit_LR"},
+        "init_pg_lrs": [1.0e-06, 2.0e-06],
+    }
+    fsdp_reinitlr_optim_sched_dict = deepcopy(fsdp_reinitlr_sched_dict)
+    fsdp_reinitlr_optim_sched_dict[1]["new_optimizer"] = deepcopy(fsdp_reinit_optim_sched_dict[1]["new_optimizer"])
+    fsdp_reinitlr_optim_sched_dict[2]["new_optimizer"] = deepcopy(fsdp_reinit_optim_sched_dict[2]["new_optimizer"])
+    fsdp_reinitlr_optim_sched_dict[2]["new_lr_scheduler"] = {
+        "lr_scheduler_init": {
+            "class_path": "torch.optim.lr_scheduler.StepLR",
+            "init_args": {"step_size": 1, "gamma": 0.2, "verbose": True},
+        },
+        "pl_lrs_cfg": {"interval": "epoch", "frequency": 1, "name": "Custom_Reinit_LR"},
+    }
     fsdp_bn_gen_sched_dict = deepcopy(fsdp_gen_sched_dict)
     fsdp_bn_gen_sched_dict[0]["params"] = ["layer.(8|[4-6]).*"]
     fsdp_bn_gen_sched_dict[1]["params"] = ["layer.[1-3].*"]
@@ -141,16 +192,19 @@ def fsdp_ft_schedules(tmpdir_factory) -> Tuple[Path, Dict]:
         fsdp_nondis_mod_ex_sched_dict,
         fsdp_ext_gen_sched_dict,
         fsdp_epoch_only_sched,
+        fsdp_reinitlr_optim_sched_dict,
     )
 
 
-@pytest.fixture(scope="function")
-def fsdp_ckpt(tmpdir_factory, fsdp_ft_schedules) -> Dict:
+@pytest.fixture(scope="module", params=[True, False], ids=["use_orig", "no_use_orig"])
+def fsdp_ckpt(tmpdir_factory, fsdp_ft_schedules, request) -> Tuple[Dict, bool]:
     """A fixture that generates a checkpoint with a sharded model."""
     seed_everything(42)
     test_model_cfg = {"fsdp_mask": {"wrapped_mods": list(range(6)), "unwrapped_mods": [7]}}
+    use_orig_kwargs = DISABLE_USE_ORIG if not request.param else {}
     strategy = FSDPStrategy(
         auto_wrap_policy=custom_auto_wrap_policy,
+        **use_orig_kwargs,
         cpu_offload=CPUOffload(offload_params=True),
     )
     callbacks = [
@@ -158,7 +212,7 @@ def fsdp_ckpt(tmpdir_factory, fsdp_ft_schedules) -> Dict:
         FTSEarlyStopping(monitor="val_loss", patience=1),
         FTSCheckpoint(monitor="val_loss", save_last=True, verbose=True),
     ]
-    model = FTSBaseFSDPModel(**test_model_cfg)
+    model = FTSAdamFSDPModel(**test_model_cfg)
     trainer = Trainer(
         default_root_dir=tmpdir_factory.getbasetemp(),
         accelerator="gpu",
@@ -168,7 +222,7 @@ def fsdp_ckpt(tmpdir_factory, fsdp_ft_schedules) -> Dict:
         max_epochs=1,
     )
     trainer.fit(model)
-    return trainer.checkpoint_callback.best_model_path
+    return trainer.checkpoint_callback.best_model_path, request.param
 
 
 ########################
@@ -198,7 +252,6 @@ class FTSBaseFSDPModel(FinetuningSchedulerBoringModel):
 
     def configure_optimizers(self):
         parameters = filter(lambda x: x.requires_grad, self.parameters())
-        # return torch.optim.SGD(parameters, lr=0.1)
         optimizer = torch.optim.SGD(parameters, lr=0.1)
         lr_scheduler = torch.optim.lr_scheduler.StepLR(optimizer, step_size=1, gamma=0.7)
         return [optimizer], [lr_scheduler]
@@ -237,8 +290,8 @@ class FTSBaseFSDPModel(FinetuningSchedulerBoringModel):
         else:
             assert isinstance(self.layer, torch.nn.Sequential)
         if self.precision_key == "auto_16":
-            assert isinstance(self.trainer.strategy.precision_plugin, FSDPMixedPrecisionPlugin)
-            precision = torch.float16 if self.trainer.precision == "16-mixed" else torch.bfloat16
+            assert isinstance(self.trainer.strategy.precision_plugin, FSDPPrecisionPlugin)
+            precision = torch.float16 if self.trainer.precision == "16-true" else torch.bfloat16
         # ensure our ignored module is not wrapped
         for i in self.fsdp_mask["unwrapped_mods"]:
             assert not isinstance(self.layer[i], FullyShardedDataParallel)
@@ -282,11 +335,11 @@ class NonDynamicLossAdamFSDPModel(FTSBaseFSDPModel):
         return [optimizer], [lr_scheduler]
 
 
-class FTSCsmFSDPModel(FTSBaseFSDPModel):
+class FTSCmFSDPModel(FTSBaseFSDPModel):
     def __init__(self, *args, **kwargs):
         super().__init__(*args, **kwargs)
 
-    def configure_sharded_model(self) -> None:
+    def configure_model(self) -> None:
         # the model is already wrapped with FSDP: no need to wrap again!
         if isinstance(self.layer, FullyShardedDataParallel):
             return
@@ -330,7 +383,7 @@ class FTSEnforceP0FSDPModel(FTSBaseFSDPModel):
         return [optimizer], [lr_scheduler]
 
 
-class FTSCsmAdamFSDPModel(FTSBaseFSDPModel):
+class FTSCmAdamFSDPModel(FTSBaseFSDPModel):
     def __init__(self, *args, **kwargs):
         super().__init__(*args, **kwargs)
 
@@ -340,7 +393,7 @@ class FTSCsmAdamFSDPModel(FTSBaseFSDPModel):
         lr_scheduler = torch.optim.lr_scheduler.StepLR(optimizer, step_size=1, gamma=0.7)
         return [optimizer], [lr_scheduler]
 
-    def configure_sharded_model(self) -> None:
+    def configure_model(self) -> None:
         # the model is already wrapped with FSDP: no need to wrap again!
         if isinstance(self.layer, FullyShardedDataParallel):
             return
@@ -350,6 +403,19 @@ class FTSCsmAdamFSDPModel(FTSBaseFSDPModel):
         self.layer = wrap(self.layer)
 
 
+class FTSAdamFSDPModel(FTSBaseFSDPModel):
+    # use of this non-SGD optimizer config is required for tests that restore optimizer state with PyTorch 2.0.x
+    # due to https://github.com/pytorch/pytorch/issues/99079
+    def __init__(self, *args, **kwargs):
+        super().__init__(*args, **kwargs)
+
+    def configure_optimizers(self):
+        parameters = filter(lambda x: x.requires_grad, self.parameters())
+        optimizer = torch.optim.AdamW(parameters, weight_decay=1.0e-05, eps=1.0e-07, lr=1.0e-05)
+        lr_scheduler = torch.optim.lr_scheduler.StepLR(optimizer, step_size=1, gamma=0.7)
+        return [optimizer], [lr_scheduler]
+
+
 class FTSExtFSDPModel(FTSBaseFSDPModel):
     def __init__(self, *args, **kwargs):
         super().__init__(*args, **kwargs)
@@ -411,6 +477,7 @@ class FSDPTestFinetuningScheduler(TestFinetuningScheduler):
 
     def restore_best_ckpt(self) -> None:
         super(TestFinetuningScheduler, self).restore_best_ckpt()
+        self.restored_best_cnt += 1
 
     def on_train_epoch_start(self, trainer, pl_module):
         super(TestFinetuningScheduler, self).on_train_epoch_start(trainer, pl_module)
@@ -419,19 +486,35 @@ class FSDPTestFinetuningScheduler(TestFinetuningScheduler):
             len(self._fts_state._curr_thawed_params),
             len(self.strategy_adapter.logical_param_translation(self._fts_state._curr_thawed_params)),
         )
-        if self.expected_state:
-            assert current_state == self.expected_state[state_key]
+        lrs_state = tuple(round(pg["lr"], 9) for pg in trainer.optimizers[0].param_groups)
+        self.inspect_or_assert(current_state, lrs_state, state_key)
+
+
+class FSDPOptInspectFTS(FSDPTestFinetuningScheduler):
+    def on_train_epoch_start(self, trainer, pl_module):
+        super(TestFinetuningScheduler, self).on_train_epoch_start(trainer, pl_module)
+        state_key = trainer.current_epoch
+        current_state = (
+            len(self._fts_state._curr_thawed_params),
+            len(self.strategy_adapter.logical_param_translation(self._fts_state._curr_thawed_params)),
+            trainer.optimizers[0].__class__.__name__,
+            trainer.fit_loop.epoch_loop.automatic_optimization.optim_progress.optimizer_steps,
+            trainer.optimizers[0].defaults["lr"],
+        )
+        lrs_state = tuple(round(pg["lr"], 9) for pg in trainer.optimizers[0].param_groups)
+        self.inspect_or_assert(current_state, lrs_state, state_key)
 
 
 # model aliases
 base_model = FTSBaseFSDPModel
 nond_loss_adam_model = NonDynamicLossAdamFSDPModel
-cust_model = FTSCsmFSDPModel
+cust_model = FTSCmFSDPModel
 wrapped_model = AlreadyWrappedFSDPModel
 nodecay_model = FTSNoDecayFSDPModel
 BN_model = FTSBatchNormFSDPModel
 shared_model = FTSSharedParamFSDPModel
-csm_adam_model = FTSCsmAdamFSDPModel
+cm_adam_model = FTSCmAdamFSDPModel
+adam_model = FTSAdamFSDPModel
 ext_model = FTSExtFSDPModel
 enforceP0_model = FTSEnforceP0FSDPModel
 
@@ -444,6 +527,7 @@ unwrap_5_7 = {"fsdp_mask": {"wrapped_mods": list(range(5)), "unwrapped_mods": [5
 unwrap_0_1_7 = {"fsdp_mask": {"wrapped_mods": [2, 3, 4, 5], "unwrapped_mods": [0, 1, 7]}}
 wrap_5_7 = {"fsdp_mask": {"wrapped_mods": [5, 7], "unwrapped_mods": [i for i in list(range(8)) if i not in [5, 7]]}}
 wrap_5 = {"fsdp_mask": {"wrapped_mods": [5], "unwrapped_mods": [i for i in list(range(8)) if i != 5]}}
+outer_wrap_only = {"fsdp_mask": {"wrapped_mods": [], "unwrapped_mods": list(range(8))}}
 unwrap_7_diverge = {"fsdp_mask": {"wrapped_mods": list(range(6)), "unwrapped_mods": [7]}, "diverge_on_epoch": 1}
 unwrap_7_mp = {"fsdp_mask": {"wrapped_mods": list(range(6)), "unwrapped_mods": [7]}, **fp16_cfg}
 unwrap_8_mp = {"fsdp_mask": {"wrapped_mods": list(range(7)), "unwrapped_mods": [8]}, **fp16_cfg}
@@ -490,11 +574,28 @@ class CustomWrapPolicy(_FSDPPolicy):
         return self._policy
 
 
+# RunIf aliases
+runif_map = {
+    "min2_0": {"min_torch": "2.0.0"},
+    "only2_0": {"min_torch": "2.0.0", "max_torch": "2.0.1"},
+    "min2_1": {"min_torch": "2.1.0"},
+    "max1_13": {"max_torch": "1.13.1"},
+    "min2_0_max3_11": {"min_torch": "2.0.0", "max_python": "3.11"},
+}
+
 # auto-wrap policy aliases
 cust_awp = custom_auto_wrap_policy
 cust_ext_awp = custom_auto_wrap_ext_policy
 warn_cust_awp = warn_custom_auto_wrap_policy
-awp_mwp_parity = CustomWrapPolicy(min_num_params=67)
+if _TORCH_GREATER_EQUAL_2_1:
+    numel_constant = 67
+    awp_mwp_2_1_parity = CustomPolicy(lambda_fn=lambda m: sum(p.numel() for p in m.parameters()) >= numel_constant)
+    awp_mwp_2_0_parity = None
+elif _TORCH_GREATER_EQUAL_2_0:
+    awp_mwp_2_0_parity = CustomWrapPolicy(min_num_params=67)
+    awp_mwp_2_1_parity = None
+else:
+    awp_mwp_2_0_parity, awp_mwp_2_1_parity = None, None
 
 # awp_overrides configuration aliases
 awp_5_9 = {"awp_overrides": ["layer.9", "layer.5"]}
@@ -503,25 +604,30 @@ awp_7 = {"awp_overrides": ["layer.7"]}
 awp_7_8 = {"awp_overrides": ["l.*yer.8", "layer.7"]}
 
 # FSDP strategy configuration aliases
-act_ckpt_cfg = {"activation_checkpointing": [torch.nn.Linear]}
-ignore_mod_cfg = {"test_ignored_modules_names": ["layer.4"], "cpu_offload": False}
-ignore_params_uo_cfg = {
+act_ckpt_cfg = {"activation_checkpointing_policy": {torch.nn.Linear}, **DISABLE_USE_ORIG}
+ignore_mod_cfg = {"test_ignored_modules_names": ["layer.4"], "cpu_offload": False, **DISABLE_USE_ORIG}
+ignore_states_cfg = {
     "test_ignored_parameters_names": ["layer.4.weight", "layer.4.bias"],
     "cpu_offload": False,
     "use_orig_params": True,
 }
-test_use_orig = {"use_orig_params": True}
+
+if not _TORCH_GREATER_EQUAL_2_0:
+    cust_mp_args = {"param_dtype": torch.float16, "reduce_dtype": torch.float16, "buffer_dtype": torch.float16}
+cust_mp_args = {"param_dtype": torch.float16, "reduce_dtype": torch.float16, "buffer_dtype": torch.float16}
+cust_fp16_mp = {"mixed_precision": MixedPrecision(**cust_mp_args), **DISABLE_USE_ORIG} if _min_fsdp_available else {}
 
 # trainer configuration alias
 max_epoch_5 = {"max_epochs": 5}
 max_epoch_4 = {"max_epochs": 4}
 
-# fts config alias
+# fts config aliases
 epoch_t_only = {
     "epoch_transitions_only": True,
     "test_es": "disable",
     "test_ckpt": ExplicitLossFTSCheckpoint(monitor="val_loss", verbose=True),
 }
+opt_inspect = {"test_fts": FSDPOptInspectFTS}
 
 # expected training path aliases
 path_default = {0: (2, 4), 1: (6, 12), 2: (7, 14)}
@@ -533,154 +639,198 @@ path_8_16 = {0: (4, 8), 1: (7, 14), 2: (8, 16)}
 path_5_10 = {0: (2, 4), 1: (3, 6), 2: (5, 10)}
 path_ext_7_14 = {0: (2, 4), 1: (2, 4), 2: (6, 12), 3: (6, 12), 4: (7, 14)}
 path_ext_8_16 = {0: (3, 6), 1: (7, 14), 2: (8, 16)}
+path_optimlr_reinit = {0: (2, 4, "SGD", 0, 0.1), 1: (6, 12, "Adam", 32, 0.00021), 2: (7, 14, "SGD", 64, 0.002)}
+lrs_path_default = {0: (0.1,), 1: (0.07, 1e-06), 2: (0.049, 7e-07, 1e-05)}
+lrs_path_optimlr_reinit = {0: (0.1,), 1: (0.00021, 1e-06), 2: (0.002, 1e-06, 3e-06)}
 
-
-# to help dedup config
-def nones(num_n) -> Tuple:
-    return (None,) * num_n
-
-
-EXPECTED_FSDP_FTS_RESULTS = {
-    "cust_awp_noprec": (path_default, *nones(2)),
-    "cust_awp_noprec_use_orig": (path_default_orig, *nones(2)),
-    "cust_awp_noprec_dynamo_use_orig": (path_default_orig_eo_dyn, *nones(2)),
-    "cust_awp_mwp_parity": (path_default, *nones(2)),
-    "override_csm_noprec": (path_default, *nones(2)),
-    # TODO: once PyTorch deprecates ``ignored_modules``, check for that deprecation warning in this test
-    "cust_awp_nop_ignore_m_no_ofld": (path_8_14, *nones(2)),
-    "cust_awp_nop_ignore_p_no_ofld_uo": (path_ignore_p_uo, *nones(2)),
-    "unsupp_torch_version": ({}, None, "is supported from PyTorch"),
-    "non_disjoint_phase_fsdp_params": ({}, None, "do not have disjoint FSDP-flattened parameter"),
-    "non_disjoint_phase_mods": ({}, None, "not have disjoint"),
-    "non_disjoint_excluded_ft_params": ({}, None, "parameters not included in"),
-    "already_fsdp_wrapped": ({}, None, "already wrapped by FSDP"),
-    "no_fsdp_params_p0": ({}, None, "one or more FSDP"),
-    "warn_unsupp_nodecay": ({}, "will now be unset", None),
-    "unmatched_awp_overrides": ({}, None, "did not match any named modules"),
-    "cust_awp_prec": (path_default, *nones(2)),
-    "cust_awp_prec_pt1x": (path_default, *nones(2)),
-    "enforceP0_cust_awp_prec": (path_default, *nones(2)),
-    # "batch_norm_auto_prec": (path_8_16, "Both mixed precision", None),  # _dynamo/allowed_functions.py suppresses
-    "batch_norm_auto_prec": (path_8_16, None, None),
-    "shared_params_auto_prec": (path_5_10, ("Pruning explicitly specified",), None),
-    "override_csm_adam_noprec": (path_ext_7_14, *nones(2)),
-    "cust_awp_overrides_prec": (path_default, *nones(2)),
-    "cust_awp_overrides_prec_ext": (path_ext_8_16, *nones(2)),
-    "warn_ignore_awp_override": ({}, "will be unset and not applied", None),
-    "cust_noprec_resume": (path_default, *nones(2)),
+# consolidate all core FTS FSDP test configuration into this dictionary to dedup config
+FTS_FSDP_TESTS = {
+    "cust_awp_noprec_no_use_orig": (
+        (adam_model, cust_awp, False, 0, unwrap_7, *nones(3), act_ckpt_cfg),
+        None,
+        (path_default, *nones(3)),
+    ),
+    "cust_awp_noprec": (
+        (adam_model, cust_awp, False, 0, unwrap_7, *nones(4)),
+        "min2_0",
+        (path_default_orig, *nones(3)),
+    ),
+    "noawp_noprec": (
+        (base_model, None, False, 0, outer_wrap_only, *nones(4)),
+        "min2_1",
+        (path_default_orig, "degenerate and unintended", *nones(2)),
+    ),
+    "noawp_prec": (
+        (base_model, None, True, 0, outer_wrap_only, *nones(4)),
+        "min2_1",
+        (path_default_orig, "degenerate and unintended", *nones(2)),
+    ),
+    "cust_awp_noprec_dynamo": (
+        (nond_loss_adam_model, cust_awp, False, 7, unwrap_7_dyn, None, epoch_t_only, max_epoch_4, None),
+        "min2_0_max3_11",
+        (path_default_orig_eo_dyn, *nones(3)),
+    ),
+    "cust_awp_mwp_2_1_reinitlr_optim_no_use_orig": (
+        (base_model, awp_mwp_2_1_parity, True, 8, unwrap_7_mp, None, opt_inspect, None, DISABLE_USE_ORIG),
+        "min2_1",
+        (
+            path_optimlr_reinit,
+            ("Incompatible check",),
+            None,
+            lrs_path_optimlr_reinit,
+        ),
+    ),
+    "cust_awp_mwp_2_0_parity_no_use_orig": (
+        (adam_model, awp_mwp_2_0_parity, True, 0, unwrap_7_mp, *nones(3), DISABLE_USE_ORIG),
+        "only2_0",
+        (path_default, *nones(3)),
+    ),
+    "cust_awp_mwp_2_1_parity_no_use_orig": (
+        (base_model, awp_mwp_2_1_parity, True, 0, unwrap_7_mp, *nones(3), DISABLE_USE_ORIG),
+        "min2_1",
+        (path_default, *nones(3)),
+    ),
+    "override_cm_noprec_no_use_orig": (
+        (cust_model, None, False, 0, unwrap_7, *nones(3), DISABLE_USE_ORIG),
+        "min2_1",
+        (path_default, *nones(3)),
+    ),
+    "cust_awp_nop_ignore_m_no_ofld_no_use_orig": (
+        (base_model, cust_awp, False, 0, unwrap_4_7, *nones(3), ignore_mod_cfg),
+        "min2_1",
+        (path_8_14, *nones(3)),
+    ),  # TODO: once PyTorch deprecates ``ignored_modules``, check for the warning with this test
+    "cust_awp_nop_ignore_p_no_ofld": (
+        (base_model, cust_awp, False, 0, unwrap_4_7, *nones(3), ignore_states_cfg),
+        "min2_1",
+        (path_ignore_p_uo, *nones(3)),
+    ),
+    "unsupp_torch_version": (
+        (base_model, cust_awp, False, 0, unwrap_7, *nones(4)),
+        None,
+        ({}, None, "is supported from PyTorch", None),
+    ),
+    "non_disjoint_params_allowed": (
+        (base_model, warn_cust_awp, False, 0, outer_wrap_only, *nones(4)),
+        "min2_1",
+        (path_default_orig, *nones(3)),
+    ),
+    "non_disjoint_phase_fsdp_params_no_use_orig": (
+        (base_model, warn_cust_awp, False, 0, wrap_5, *nones(3), DISABLE_USE_ORIG),
+        None,
+        ({}, None, "do not have disjoint FSDP-flattened parameter", None),
+    ),
+    "non_disjoint_phase_mods_no_use_orig": (
+        (cust_model, None, False, 1, unwrap_7, *nones(3), DISABLE_USE_ORIG),
+        None,
+        ({}, None, "not have disjoint", None),
+    ),
+    "non_disjoint_excluded_ft_params_no_use_orig": (
+        (cust_model, None, False, 5, unwrap_0_1_7, *nones(3), DISABLE_USE_ORIG),
+        None,
+        ({}, None, "parameters not included in", None),
+    ),
+    "already_fsdp_wrapped": (
+        (wrapped_model, cust_awp, False, 0, unwrap_7, *nones(4)),
+        "min2_1",
+        ({}, None, "already wrapped by FSDP", None),
+    ),
+    "no_fsdp_params_p0": (
+        (cust_model, None, False, 0, unwrap_5_7, *nones(3), DISABLE_USE_ORIG),
+        None,
+        ({}, None, "one or more FSDP", None),
+    ),
+    "no_nonzero_local_shards_p0": (
+        (base_model, warn_cust_awp, True, 0, outer_wrap_only, *nones(4)),
+        "min2_1",
+        (path_default_orig, *nones(3)),
+    ),  # exercise shard allocation DEBUG diagnostics
+    "warn_unsupp_nodecay": (
+        (nodecay_model, cust_awp, False, 0, unwrap_7, *nones(4)),
+        "min2_1",
+        ({}, "will now be unset", *nones(2)),
+    ),
+    "unmatched_awp_overrides": (
+        (base_model, warn_cust_awp, True, 0, wrap_5_7, awp_5_9, *nones(3)),
+        None,
+        ({}, None, "did not match any named modules", None),
+    ),
+    "cust_awp_prec_no_use_orig": (
+        (base_model, cust_awp, True, 0, unwrap_7_mp, *nones(3), DISABLE_USE_ORIG),
+        "min2_1",
+        (path_default, *nones(3)),
+    ),
+    "cust_awp_prec_pt1x": (
+        (base_model, cust_awp, True, 0, unwrap_7_mp, *nones(3), cust_fp16_mp),
+        "max1_13",
+        (path_default, *nones(3)),
+    ),
+    "enforceP0_cust_awp_prec_no_use_orig": (
+        (enforceP0_model, cust_awp, True, 0, unwrap_7_mp, *nones(3), DISABLE_USE_ORIG),
+        "min2_1",
+        (path_default, *nones(3)),
+    ),
+    "batch_norm_auto_prec_no_use_orig": (
+        (BN_model, cust_awp, True, 2, unwrap_8_mp, *nones(3), DISABLE_USE_ORIG),
+        "min2_1",
+        (path_8_16, ("Both mixed precision",), *nones(2)),
+    ),
+    "shared_params_auto_prec_no_use_orig": (
+        (shared_model, cust_awp, True, 3, unwrap_7_mp, awp_1, *nones(2), DISABLE_USE_ORIG),
+        "min2_1",
+        (
+            path_5_10,
+            ("Pruning explicitly specified",),
+            *nones(2),
+        ),
+    ),
+    "override_cm_adam_noprec_no_use_orig": (
+        (cm_adam_model, None, False, 4, unwrap_7_diverge, *nones(2), max_epoch_5, DISABLE_USE_ORIG),
+        None,
+        (path_ext_7_14, *nones(3)),
+    ),
+    "cust_awp_overrides_prec_no_use_orig": (
+        (adam_model, cust_awp, True, 0, wrap_all_mp, awp_7, *nones(2), cust_fp16_mp),
+        "min2_0",
+        (path_default, *nones(3)),
+    ),
+    "cust_awp_overrides_mwp_prec_no_use_orig": (
+        (base_model, awp_mwp_2_1_parity, True, 0, wrap_all_mp, awp_7, *nones(2), cust_fp16_mp),
+        "min2_1",
+        (path_default, *nones(3)),
+    ),
+    "cust_awp_overrides_mwp_prec_ext_no_use_orig": (
+        (ext_model, cust_ext_awp, True, 6, wrap_ext_mp, awp_7_8, *nones(2), cust_fp16_mp),
+        "min2_1",
+        (path_ext_8_16, *nones(3)),
+    ),
+    "warn_ignore_awp_override": (
+        (cust_model, None, False, 0, unwrap_7, awp_7, *nones(3)),
+        "min2_1",
+        ({}, "will be unset and not applied", *nones(2)),
+    ),
 }
 
 
+EXPECTED_FSDP_FTS_RESULTS = {k: test_cfg[2] for k, test_cfg in FTS_FSDP_TESTS.items()}
+
+
+FSDP_TEST_CFGS = [
+    pytest.param(
+        k,
+        *test_cfg[0],
+        id=k,
+        marks=RunIf(**runif_map[test_cfg[1]]) if runif_map.get(test_cfg[1], None) else tuple(),
+    )
+    for k, test_cfg in FTS_FSDP_TESTS.items()
+]
+
+
 @RunIf(min_cuda_gpus=2, skip_windows=True, standalone=True, min_torch="1.13")
 @pytest.mark.parametrize(
     "model_cfg_key, model_cls, auto_wrap_policy, use_precision, ft_sched_idx, model_cfg, strategy_adapter_cfg, fts_cfg,\
           trainer_cfg, strategy_cfg",
-    [
-        ("cust_awp_noprec", base_model, cust_awp, False, 0, unwrap_7, *nones(3), act_ckpt_cfg),
-        pytest.param(
-            "cust_awp_noprec_use_orig",
-            base_model,
-            cust_awp,
-            False,
-            0,
-            unwrap_7,
-            *nones(3),
-            test_use_orig,
-            marks=RunIf(min_torch="2.0"),
-        ),
-        pytest.param(
-            "cust_awp_noprec_dynamo_use_orig",
-            nond_loss_adam_model,
-            cust_awp,
-            False,
-            7,
-            unwrap_7_dyn,
-            None,
-            epoch_t_only,
-            max_epoch_4,
-            test_use_orig,
-            marks=RunIf(min_torch="2.0"),
-        ),
-        pytest.param(
-            "cust_awp_mwp_parity",
-            base_model,
-            awp_mwp_parity,
-            True,
-            0,
-            unwrap_7_mp,
-            *nones(4),
-            marks=RunIf(min_torch="2.0"),
-        ),
-        ("override_csm_noprec", cust_model, None, False, 0, unwrap_7, *nones(4)),
-        ("cust_awp_nop_ignore_m_no_ofld", base_model, cust_awp, False, 0, unwrap_4_7, *nones(3), ignore_mod_cfg),
-        pytest.param(
-            "cust_awp_nop_ignore_p_no_ofld_uo",
-            base_model,
-            cust_awp,
-            False,
-            0,
-            unwrap_4_7,
-            *nones(3),
-            ignore_params_uo_cfg,
-            marks=RunIf(min_torch="2.0"),
-        ),
-        ("unsupp_torch_version", base_model, cust_awp, False, 0, unwrap_7, *nones(4)),
-        ("non_disjoint_phase_fsdp_params", base_model, warn_cust_awp, False, 0, wrap_5, *nones(4)),
-        ("non_disjoint_phase_mods", cust_model, None, False, 1, unwrap_7, *nones(4)),
-        ("non_disjoint_excluded_ft_params", cust_model, None, False, 5, unwrap_0_1_7, *nones(4)),
-        pytest.param(
-            "already_fsdp_wrapped",
-            wrapped_model,
-            cust_awp,
-            False,
-            0,
-            unwrap_7,
-            *nones(4),
-            marks=RunIf(min_torch="2.0.0"),
-        ),
-        ("no_fsdp_params_p0", cust_model, None, False, 0, unwrap_5_7, *nones(4)),
-        ("warn_unsupp_nodecay", nodecay_model, cust_awp, False, 0, unwrap_7, *nones(4)),
-        ("unmatched_awp_overrides", base_model, warn_cust_awp, True, 0, wrap_5_7, awp_5_9, *nones(3)),
-        ("cust_awp_prec", base_model, cust_awp, True, 0, unwrap_7_mp, *nones(4)),
-        pytest.param(
-            "cust_awp_prec_pt1x", base_model, cust_awp, True, 0, unwrap_7_mp, *nones(4), marks=RunIf(max_torch="2.0.0")
-        ),
-        ("enforceP0_cust_awp_prec", enforceP0_model, cust_awp, True, 0, unwrap_7_mp, *nones(4)),
-        pytest.param(
-            "batch_norm_auto_prec", BN_model, cust_awp, True, 2, unwrap_8_mp, *nones(4), marks=RunIf(min_torch="2.0.0")
-        ),
-        ("shared_params_auto_prec", shared_model, cust_awp, True, 3, unwrap_7_mp, awp_1, *nones(3)),
-        ("override_csm_adam_noprec", csm_adam_model, None, False, 4, unwrap_7_diverge, *nones(2), max_epoch_5, None),
-        ("cust_awp_overrides_prec", base_model, cust_awp, True, 0, wrap_all_mp, awp_7, *nones(3)),
-        ("cust_awp_overrides_prec_ext", ext_model, cust_ext_awp, True, 6, wrap_ext_mp, awp_7_8, *nones(3)),
-        ("warn_ignore_awp_override", cust_model, None, False, 0, unwrap_7, awp_7, *nones(3)),
-    ],
-    ids=[
-        "cust_awp_noprec",
-        "cust_awp_noprec_use_orig",
-        "cust_awp_noprec_dynamo_use_orig",
-        "cust_awp_mwp_parity",
-        "override_csm_noprec",
-        "cust_awp_nop_ignore_m_no_ofld",
-        "cust_awp_nop_ignore_p_no_ofld_uo",
-        "unsupp_torch_version",
-        "non_disjoint_phase_fsdp_params",
-        "non_disjoint_phase_mods",
-        "non_disjoint_excluded_ft_params",
-        "already_fsdp_wrapped",
-        "no_fsdp_params_p0",
-        "warn_unsupp_nodecay",
-        "unmatched_awp_overrides",
-        "cust_awp_prec",
-        "cust_awp_prec_pt1x",
-        "enforceP0_cust_awp_prec",
-        "batch_norm_auto_prec",
-        "shared_params_auto_prec",
-        "override_csm_adam_noprec",
-        "cust_awp_overrides_prec",
-        "cust_awp_overrides_prec_ext",
-        "warn_ignore_awp_override",
-    ],
+    FSDP_TEST_CFGS,
 )
 def test_fsdp_multi_gpus(
     tmpdir,
@@ -699,46 +849,45 @@ def test_fsdp_multi_gpus(
 ):
     """Conservative (end-to-end) set of tests for FTS support of FSDP."""
     cfg = init_test_cfg(model_cfg_key, model_cfg, fts_cfg, trainer_cfg, strategy_cfg, use_precision)
-    fts_state, warns_expected, exception_expected, model_cfg, fts_cfg, trainer_cfg, strategy_cfg, precision_opts = cfg
+    (fts_state, lrs_state, w_expected, e_expected, model_cfg, fts_cfg, trainer_cfg, strategy_cfg, precision_opts) = cfg
     seed_everything(42)
     use_dynamo = True if model_cfg.pop("use_dynamo", None) else False
     model = model_cls(**model_cfg)
     strategy_cfg = load_ignore_directives(strategy_cfg, model)
     ft_sched = fsdp_ft_schedules[ft_sched_idx]
-    test_cfg = init_fts_cfg(fts_state, strategy_adapter_cfg, fts_cfg)
-    callbacks = callbacks_cfg(FSDPTestFinetuningScheduler, ft_sched, test_cfg, {"patience": 2}, {"save_top_k": 3})
+    test_cfg, fts_cls = init_fts_cfg(fts_state, lrs_state, strategy_adapter_cfg, fts_cfg, tmpdir)
+    callbacks = callbacks_cfg(fts_cls, ft_sched, test_cfg, {"patience": 2}, {"save_top_k": 3})
     strategy = FSDPStrategy(auto_wrap_policy=auto_wrap_policy, **strategy_cfg)
     trainer = configure_trainer(tmpdir, strategy, callbacks, {**trainer_cfg, **precision_opts})
     configured_model = torch.compile(model) if use_dynamo else model
-    if exception_expected:
-        gen_exceptions(trainer, configured_model, model_cfg_key, exception_expected)
+    if e_expected:
+        gen_exceptions(trainer, configured_model, model_cfg_key, e_expected)
     else:
         trainer.fit(configured_model)
         default_fts_sanity_chk(trainer)
-    if trainer.is_global_zero:
-        check_fts_fsdp_warns(warns_expected, recwarn, use_dynamo)
+    if trainer.is_global_zero and recwarn:
+        check_fts_fsdp_warns(w_expected, recwarn, use_dynamo)
 
 
 @RunIf(min_cuda_gpus=2, skip_windows=True, standalone=True, min_torch="1.13")
 @pytest.mark.parametrize(
-    "model_cfg_key, model_cls, awp, ft_sched_idx, model_cfg",
-    [("cust_noprec_resume", base_model, cust_awp, 0, unwrap_7)],
-    ids=["cust_noprec_resume"],
+    "model_cls, awp, ft_sched_idx, model_cfg",
+    [pytest.param(adam_model, cust_awp, 0, unwrap_7, id="cust_noprec_resume")],
 )
-def test_fsdp_multi_gpus_resume(
-    tmpdir, recwarn, fsdp_ft_schedules, fsdp_ckpt, model_cfg_key, model_cls, awp, ft_sched_idx, model_cfg
-):
+def test_fsdp_multi_gpus_resume(tmpdir, recwarn, fsdp_ft_schedules, fsdp_ckpt, model_cls, awp, ft_sched_idx, model_cfg):
     """Conservative (end-to-end) test for FTS training resumption with FSDP."""
     model_cfg = model_cfg or {}
-    expected_state = EXPECTED_FSDP_FTS_RESULTS[model_cfg_key]
+    expected_state = (path_default, *nones(3))
     warns_expected = expected_state[1]
     seed_everything(42)
     model = model_cls(**model_cfg)
     ft_sched = fsdp_ft_schedules[ft_sched_idx]
     callbacks = callbacks_cfg(FinetuningScheduler, ft_sched, {}, {"patience": 1}, {"save_last": True})
-    strategy = FSDPStrategy(auto_wrap_policy=awp, cpu_offload=CPUOffload(offload_params=True))
+    ckpt_path, use_orig_params = fsdp_ckpt
+    use_orig_kwargs = DISABLE_USE_ORIG if not use_orig_params else {}
+    strategy = FSDPStrategy(auto_wrap_policy=awp, **use_orig_kwargs, cpu_offload=CPUOffload(offload_params=True))
     trainer = configure_trainer(tmpdir, strategy, callbacks, {"max_epochs": 3})
-    trainer.ckpt_path = fsdp_ckpt
+    trainer.ckpt_path = ckpt_path
     trainer.fit(model)
     default_fts_sanity_chk(trainer)
     if trainer.is_global_zero:
@@ -747,7 +896,7 @@ def test_fsdp_multi_gpus_resume(
 
 def gen_exceptions(trainer, model, model_cfg_key, exception_expected):
     if model_cfg_key == "no_fsdp_params_p0":
-        with mock.patch.object(FSDPStrategyAdapter, "RANK_ZERO_LOG_FQN", 42):
+        with mock.patch.object(FSDPStrategyAdapter, "_rank_zero_logger", 42):
             with pytest.raises(MisconfigurationException, match=exception_expected):
                 trainer.fit(model)
     elif model_cfg_key == "unsupp_torch_version":
@@ -759,18 +908,35 @@ def gen_exceptions(trainer, model, model_cfg_key, exception_expected):
             trainer.fit(model)
 
 
-def init_fts_cfg(fts_state, strategy_adapter_cfg, fts_cfg):
-    def_fts_cfg = {"logging_level": DEBUG, "expected_state": fts_state, "strategy_adapter_cfg": strategy_adapter_cfg}
+def init_fts_cfg(fts_state, lrs_state, strategy_adapter_cfg, fts_cfg, tmpdir):
+    def_fts_cfg = {
+        "logging_level": DEBUG,
+        "expected_state": fts_state,
+        "lrs_state": lrs_state,
+        "strategy_adapter_cfg": strategy_adapter_cfg,
+        # "state_log_dir": tmpdir
+    }
+    fts_cls = fts_cfg.pop("test_fts") if fts_cfg and fts_cfg.get("test_fts") else FSDPTestFinetuningScheduler
     test_cfg = {**fts_cfg, **def_fts_cfg}
-    return test_cfg
+    return test_cfg, fts_cls
 
 
 def init_test_cfg(model_cfg_key, model_cfg, fts_cfg, trainer_cfg, strategy_cfg, use_precision):
     expected_state = EXPECTED_FSDP_FTS_RESULTS[model_cfg_key]
-    fts_state, warns_expected, exception_expected = expected_state
+    fts_state, warns_expected, exception_expected, lrs_state = expected_state
     init_cfg = model_cfg, fts_cfg, trainer_cfg, strategy_cfg, use_precision
     model_cfg, fts_cfg, trainer_cfg, strategy_cfg, precision_opts = map_component_cfgs(*init_cfg)
-    return fts_state, warns_expected, exception_expected, model_cfg, fts_cfg, trainer_cfg, strategy_cfg, precision_opts
+    return (
+        fts_state,
+        lrs_state,
+        warns_expected,
+        exception_expected,
+        model_cfg,
+        fts_cfg,
+        trainer_cfg,
+        strategy_cfg,
+        precision_opts,
+    )
 
 
 def map_component_cfgs(model_cfg, fts_cfg, trainer_cfg, strategy_cfg, use_precision):
@@ -778,13 +944,13 @@ def map_component_cfgs(model_cfg, fts_cfg, trainer_cfg, strategy_cfg, use_precis
     model_cfg = model_cfg or {}
     fts_cfg = fts_cfg or {}
     strategy_cfg = strategy_cfg or {"cpu_offload": True}
-    precision_opts = {"precision": "16-mixed"} if use_precision else {}
+    precision_opts = {"precision": "16-true"} if use_precision else {}
     return model_cfg, fts_cfg, trainer_cfg, strategy_cfg, precision_opts
 
 
 def load_ignore_directives(strategy_cfg, model):
     if strategy_cfg.get("test_ignored_parameters_names", None):
-        strategy_cfg["ignored_parameters"] = [
+        strategy_cfg["ignored_states"] = [
             model.get_parameter(n) for n in strategy_cfg.pop("test_ignored_parameters_names")
         ]
     elif strategy_cfg.get("test_ignored_modules_names", None):
